@@ -1,11 +1,11 @@
 ;;; scripts/compile-ffi.scm --- NYACC's command-line FFI compiler
 
-;; Copyright (C) 2017-2021 Matthew R. Wette
+;; Copyright (C) 2017-2021,2024 Matthew Wette
 ;;
-;; This program is free software; you can redistribute it and/or
-;; modify it under the terms of the GNU Lesser General Public
-;; License as published by the Free Software Foundation; either
-;; version 3 of the License, or (at your option) any later version.
+;; This program is free software; you can redistribute it and/or modify it
+;; under the terms of the GNU Lesser General Public License as published by
+;; the Free Software Foundation; either version 3 of the License, or (at
+;; your option) any later version.
 ;;
 ;; This library is distributed in the hope that it will be useful,
 ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -24,23 +24,25 @@
 ;;; Code:
 
 (define-module (scripts compile-ffi)
-  #:use-module (nyacc lang c99 ffi-help)
+  #:use-module (nyacc foreign arch-info)
   #:use-module (system base language)
   #:use-module ((system base compile) #:select (compile-file))
-  #:use-module ((srfi srfi-1) #:select (fold fold-right))
+  #:use-module ((srfi srfi-1) #:select (fold fold-right lset-union every))
   #:use-module (srfi srfi-37)
-  #:version (1 09 4))
-(cond-expand
- (guile-3
-  (define (compile-scm file)
-    (compile-file file #:from 'scheme #:to 'bytecode
-                  #:optimization-level 0 #:opts '())))
- (guile-2
-  (define (compile-scm file)
-    (compile-file file #:from 'scheme #:to 'bytecode
-                  #:opts '()))))
+  #:version (2 00 0))
 
-(define *ffi-help-version* "1.09.4")
+(define *ffi-help-version* "2.00.0")
+
+(define (compile-scm file)
+  (compile-file file
+                #:from 'scheme #:to 'bytecode
+                #:optimization-level 1 #:opts '()))
+
+(define (fix-path path)
+  (let* ((cwd (getcwd)))
+    (if (string-contains path cwd)
+        (substring/shared path (1+ (string-length cwd)))
+        path)))
 
 (define %summary
   "Compile a ffi-file to .scm and maybe .go.")
@@ -75,17 +77,19 @@
   (simple-format #t "Usage: compile [OPTION] FILE
 Generate a Guile Scheme file from the source FFI file FILE.
 
-  -h, --help           print this help message
-  --version            print version number
+  -h, --help            print this help message
+  --version             print version number
 
-  -L, --load-path=DIR  add DIR to the front of the module load path
-  -I, --inc-dir=DIR    add DIR to list of dir's to search for C headers
-  -o, --output=OFILE   write output to OFILE
-  -d, --debug=x,y      set debug flags: echo-decl, parse
-  -s, --show-incs      show includes during parsing
-  -D, --list-deps      list dependencies and quit
-  -X, --no-exec        don't generate .go file(s)
-  -R, --no-recurse     don't do recursive compile on dep's
+  -b, --backend=BACKEND back end target: bytestructures or cdata
+  -L, --load-path=DIR   add DIR to the front of the module load path
+  -I, --inc-dir=DIR     add DIR to list of dir's to search for C headers
+  -o, --output=OFILE    write output to OFILE
+  -d, --debug=x,y       set debug flags: echo-decl, parse
+  -s, --show-incs       show includes during parsing
+  -m, --machine=MACHINE target machine, if non-native (e.g., i686)
+  -D, --list-deps       list dependencies and quit
+  -X, --no-exec         don't generate .go file(s)
+  -R, --no-recurse      don't do recursive compile on dep's
 
 Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
 
@@ -107,6 +111,12 @@ Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
              (if (assoc-ref opts 'output-file)
                  (fail "`-o' option cannot be specified more than once"))
              (values (acons 'output arg opts) files)))
+   (option '(#\b "backend") #t #f
+           (lambda (opt name arg opts files)
+             (values (acons 'backend (string->symbol arg) opts) files)))
+   (option '(#\m "machine") #t #f
+           (lambda (opt name arg opts files)
+             (values (acons 'machine arg opts) files)))
    (option '(#\s "show-incs") #f #f
            (lambda (opt name arg opts files)
              (values (acons 'show-incs #t opts) files)))
@@ -141,42 +151,11 @@ Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
                    (string-suffix? ".ffi" file)
                    (fail "expecting .ffi suffix"))
                (values opts (cons file files)))
-             '() '()))
+             `((backend . ,(or (getenv "FFI_HELP_BACKEND") 'cdata))
+               (machine . "native")) '()))
 
 ;; --- check dependencies -------------------------------------------
 
-(define (sf fmt . args) (apply simple-format #t fmt args))
-
-(define (more-recent? ffi-file scm-file)
-  ;; copied from ice-9/boot-9.scm
-  (let ((stat1 (stat ffi-file)) (stat2 (stat scm-file)))
-    (or (> (stat:mtime stat1) (stat:mtime stat2))
-        (and (= (stat:mtime stat1) (stat:mtime stat2))
-             (>= (stat:mtimensec stat1)
-                 (stat:mtimensec stat2))))))
-
-(define (find-in-path file)
-  (let loop ((pathl %load-path))
-    (if (null? pathl) #f
-        (let ((path (string-append (car pathl) "/" file)))
-          (if (access? path R_OK) path
-              (loop (cdr pathl)))))))
-
-;; Given module spec and list of dep's, return list of out-of-date dep's.
-;; This routine assumes the scm file is in the same dir as the ffi file.
-(define (check-ffi-deps ffimod uses)
-  (fold-right
-   (lambda (fmod seed)
-     (let* ((base (string-join (map symbol->string fmod) "/"))
-            (xffi (find-in-path (string-append base ".ffi")))
-            (xscm (find-in-path (string-append base ".scm"))))
-       (unless xffi (fail "dependent ~S not found" fmod))
-       (cond
-        ((not xscm) (cons xffi seed))
-        ((more-recent? xffi xscm) (cons xffi seed))
-        (else seed))))
-   '() uses))
-
 (define-syntax find-ffi-uses
   (lambda (x)
     (syntax-case x ()
@@ -186,11 +165,45 @@ Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
       ((_ attr rest ...) #'(find-ffi-uses rest ...))
       ((_) #''()))))
 
-(define-syntax-rule (define-ffi-module path-list attr ...)
-  (check-ffi-deps (quote path-list) (find-ffi-uses attr ...)))
+(define-syntax-rule (define-ffi-module spec attr ...)
+  (cons (quote spec) (find-ffi-uses attr ...)))
 (export define-ffi-module)
 
-(define (outdated-ffi-deps file)
+(define (find-in-path file)
+  (let loop ((pathl %load-path))
+    (if (null? pathl) #f
+        (let ((path (string-append (car pathl) "/" file)))
+          (if (access? path R_OK) path
+              (loop (cdr pathl)))))))
+
+(define (more-recent? ffi-file scm-file)
+  ;; copied from ice-9/boot-9.scm
+  (let ((stat1 (stat ffi-file)) (stat2 (stat scm-file)))
+    (or (> (stat:mtime stat1) (stat:mtime stat2))
+        (and (= (stat:mtime stat1) (stat:mtime stat2))
+             (>= (stat:mtimensec stat1)
+                 (stat:mtimensec stat2))))))
+
+#;(define (stat-time<? stat1 stat2)
+  (or (< (stat:mtime stat1) (stat:mtime stat2))
+      (and (= (stat:mtime stat1) (stat:mtime stat2))
+           (< (stat:mtimensec stat1) (stat:mtimensec stat2)))))
+
+(define (tsort filed filel)
+  (define (covered? deps done) (every (lambda (e) (member e done)) deps))
+  (let loop ((done '()) (hd '()) (tl filel))
+    (if (null? tl)
+        (if (null? hd) done (loop done '() hd))
+        (cond
+         ((not (assq-ref filed (car tl)))
+          (loop (cons (car tl) done) hd (cdr tl)))
+         ((covered? (assq-ref filed (car tl)) done)
+          (loop (cons (car tl) done) hd (cdr tl)))
+         (else
+          (loop done (cons (car tl) hd) (cdr tl)))))))
+
+;; Given ffi-module filename, return alist of dependent specs
+(define (module-ffi-deps file)
   (unless (access? file R_OK)
     (fail "not found: ~S" file))
   (call-with-input-file file
@@ -203,57 +216,100 @@ Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
            ((and (pair? exp) (eqv? 'define-ffi-module (car exp))) (eval exp env))
            (else (loop (read iport)))))))))
 
-(define (gen-ffi-deps file)
-  (let loop ((odeps '()) (next '()) (prev (outdated-ffi-deps file)))
+;; from requested compiled file, return depends-on dict of .ffi-files
+(define (all-ffi-deps file)
+  ;; depd is file->mods, specd is mod->file
+  (let loop ((depd '()) (specd '()) (file file) (todo '()))
     (cond
-     ((pair? prev)
-      (loop (cons (car prev) odeps)
-            (if (member (car prev) odeps) next (cons (car prev) next))
-            (cdr prev)))
-     ((pair? next)
-      (loop odeps '() next))
+     ((and file (not (assoc-ref depd file)))
+      (let ((mdep (module-ffi-deps file))) ;; (mod-spec . dep-specs)
+        (loop (acons file (cdr mdep) depd) (acons (car mdep) file specd)
+              #f (append todo (cdr mdep)))))
+     ((pair? todo)
+      (let* ((base (string-join (map symbol->string (car todo)) "/"))
+             (path (find-in-path (string-append base ".ffi"))))
+        (unless path (fail "can't find ~s" path))
+        (loop depd specd path (cdr todo))))
      (else
       (reverse
-       (fold
-        (lambda (odep deps) (if (member odep deps) deps (cons odep deps)))
-        '() odeps))))))
+       (map (lambda (d)
+              (cons (car d) (map (lambda (m) (assoc-ref specd m)) (cdr d))))
+            depd))))))
 
-(define (ensure-ffi-deps file options)
-  (let ((options `((no-recurse . #t) (no-exec . #f) . ,options))
-        (ffi-deps (gen-ffi-deps file)))
+(define (gen-supd depd all)
+  (map (lambda (file)
+         (cons file (fold
+                     (lambda (ent s)
+                       (if (member file (cdr ent)) (cons (car ent) s) s))
+                     '() depd)))
+       all))
+
+;; from list of all files and depd, return supplier-for dict
+(define (ensure-ffi-deps file opts)
+  (define (check-depd depd seed)
+    (fold
+     (lambda (deps todo)
+       (let loop ((todo todo) (head (car deps)) (tail (cdr deps)))
+         (cond
+          ((null? tail) todo)
+          ((member head todo) todo)
+          ((more-recent? (car tail) head) (cons head todo))
+          (else (loop todo head (cdr tail))))))
+     seed depd))
+
+  (define (check-ffis ffis seed)
+    (fold
+     (lambda (ffi seed)
+       (let ((scm (find-in-path (string-append (basename ffi ".ffi") ".scm"))))
+         (if (and scm (more-recent? scm ffi)) seed
+             (if (member ffi seed) seed
+                 (cons ffi seed)))))
+     seed ffis))
+  
+  ;; 1: loop through depd.  If any cdr is more-recent add the car.  repeat
+  ;; 2: loop through all, If any ffi is more recent than ffi add it
+  ;; 3: loop through tord, for each file in todo (from 1&2) compile-ffi it
+  (let* ((depd (all-ffi-deps file))
+         (all (apply lset-union string=? depd))
+         (supd (gen-supd depd all))
+         (tord (reverse (tsort depd all)))
+         (todo (check-depd depd '()))
+         (todo (check-ffis tord todo)))
     (for-each
-     (lambda (dep-file) (compile-ffi dep-file options))
-     (reverse ffi-deps))))
+     (lambda (ffi)
+       (when (and (member ffi todo) (not (string=? ffi file)))
+         (compile-ffi ffi opts)))
+     tord)))
+
 
 ;; -----------------------------------------------------------------------------
 
-(define (fix-path path)
-  (let* ((cwd (getcwd)))
-    (if (string-contains path cwd)
-        (substring/shared path (1+ (string-length cwd)))
-        path)))
-
-(define (compile-ffi ffi-file options)
-  (let* ((base (string-drop-right ffi-file 4))
-         (scm-file (string-append base ".scm")))
-    (unless (assq-ref options 'no-recurse)
-      (ensure-ffi-deps ffi-file options))
-    (when (assq-ref options 'list-deps)
-      (for-each (lambda (f) (sfmt "~A\n" (fix-path f))) (gen-ffi-deps ffi-file))
+(define (compile-ffi ffi-file opts)
+  (let* ((scm-file (string-append (string-drop-right ffi-file 4) ".scm"))
+         (compile-ffi-file (case (assq-ref opts 'backend)
+                             ((bs bytestructures)
+                              (@ (nyacc lang c99 ffi-help-bs) compile-ffi-file))
+                             ((cd cdata)
+                              (@ (nyacc lang c99 ffi-help-cd) compile-ffi-file))
+                             (else (fail "bad backend specified")))))
+    (when (assq-ref opts 'list-deps)
+      (for-each
+       (lambda (f) (sfmt "~A\n" (fix-path f)))
+       (apply lset-union string=? (all-ffi-deps ffi-file)))
       (exit 0))
     (catch 'ffi-help-error
       (lambda ()
         (catch 'c99-error
           (lambda ()
             (sfmt "compiling `~A' ...\n" (fix-path ffi-file))
-            (compile-ffi-file ffi-file options)
+            (compile-ffi-file ffi-file opts)
             (sfmt "... wrote `~A'\n" (fix-path scm-file)))
           (lambda (key fmt . args)
             (apply throw 'ffi-help-error fmt args))))
       (lambda (key fmt . args)
         (if (access? scm-file W_OK) (delete-file scm-file))
         (apply fail fmt args)))
-    (unless (assq-ref options 'no-exec)
+    (unless (assq-ref opts 'no-exec)
       (sfmt "compiling `~A' ...\n" (fix-path scm-file))
       (let ((go-file (compile-scm scm-file)))
         (load-compiled go-file)
@@ -264,8 +320,19 @@ Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
   (call-with-values
       (lambda () (parse-args args))
     (lambda (opts files)
-      (when (or (assq-ref options 'help) (null? files)) (show-usage) (exit 0))
-      (for-each (lambda (file) (compile-ffi file opts)) (reverse files))))
+      (when (or (assq-ref opts 'help) (null? files)) (show-usage) (exit 0))
+      (unless (string=? (assq-ref opts 'machine) "native")
+        (unless (memq (assq-ref opts 'backend) '(cdata))
+          (fail "only cdata supports non-native machines architectures"))
+        (unless (getenv "CC")
+          (fail "for non-native machine architectures define env var CC")))
+      (for-each
+       (lambda (file)
+         (with-arch (assq-ref opts 'machine)
+           (unless (assq-ref opts 'no-recurse)
+             (ensure-ffi-deps file opts))
+           (compile-ffi file opts)))
+       (reverse files))))
   (exit 0))
 
 ;; --- last line ---
